from btgitserver.defaults import default_settings
from btgitserver.defaults import default_config_file_name
from btgitserver.logger import Logger
import os
from btconfig import Config

logger = Logger().init_logger(__name__)

class AppConfig():

  def __init__(self, **kwargs):
    pass

  def initialize(self, **kwargs):

    logger.info("Initializing config")
    args = kwargs.get('args', {})
    verify_tls = kwargs.get('verify_tls')
    # Initialize App Config
    initial_data = {
    'environment': os.environ
    }  

    config_file_uri = kwargs.get('config_file') or \
                      args.get('config_file') or \
                      default_config_file_name
    logger.info(f"Config file URI is {config_file_uri}")
    # Initialize App Config
    config = Config(
        config_file_uri=config_file_uri,
        default_value=default_settings,
        initial_data=initial_data,
        args=args,
        verify_tls=verify_tls
    )

    settings = config.read()

    return settings


from btgitserver.args import parse_args
from btgitserver.config import AppConfig
from btgitserver.defaults import default_app_port, \
default_app_host_address, \
default_repo_search_paths, \
default_verify_tls
from btgitserver.logger import Logger

from dulwich.pack import PackStreamReader
import subprocess
from flask_httpauth import HTTPBasicAuth
from flask import Flask, make_response, request, abort
from pathlib import Path
import sys
if sys.version_info[0] == 2:
    from StringIO import StringIO
if sys.version_info[0] >= 3:
    from io import StringIO

# Private variables
__author__ = 'berttejeda'
__original_author = 'stewartpark'

# Read command-line args
args = parse_args()
# Initialize logging facility
logger_obj = Logger(
    logfile_path=args.logfile_path,
    logfile_write_mode=args.logfile_write_mode)
logger = logger_obj.init_logger('app')

verify_tls = args.no_verify_tls or default_verify_tls

# Flask
app = Flask(__name__)
auth = HTTPBasicAuth()

# Initialize Config Readers
app_config = AppConfig().initialize(
  args=vars(args),
  verify_tls=verify_tls
)

git_search_paths = args.repo_search_paths or app_config.get('app.search_paths', default_repo_search_paths)
git_repo_map = {}

for git_search_path in git_search_paths:
  fq_git_search_path = Path(git_search_path).expanduser()
  logger.info(f'Building git repo map ...')
  logger.info(f'Adding git repos under {fq_git_search_path}')
  for p in Path(fq_git_search_path).rglob("*"):
      if p.is_dir() and p.name == '.git':
        git_directory = p.parent.as_posix()
        git_directory_name = p.parent.name
        if git_directory_name:
          git_repo_map[git_directory_name] = git_directory
logger.info(f'Finished building git repo map!')

numrepos = len(list(git_repo_map.keys()))
logger.info(f'Found {numrepos} repo(s)')

authorized_users = app_config.auth.users
users = [u[0] for u in authorized_users.items()]
available_repos = list(git_repo_map.keys())

def start_api():
  """API functions.
  This function defines the API routes and starts the API Process.
  """

  @auth.get_password
  def get_pw(username):
      if username in users:
          credential = authorized_users.get(username).password
          return credential
      else:
          return None

  @app.route('/example/<string:project_name>/info/refs')
  @app.route('/<string:project_name>/info/refs')
  @auth.login_required
  def info_refs(project_name):
      service = request.args.get('service')
      if service[:4] != 'git-':
          abort(500)

      if project_name in available_repos:
          project_path = [v for k,v in git_repo_map.items() if k == project_name][0]
          p = subprocess.Popen([service, '--stateless-rpc', '--advertise-refs', project_path], stdout=subprocess.PIPE)
          packet = '# service=%s\n' % service
          length = len(packet) + 4
          _hex = '0123456789abcdef'
          prefix = ''
          prefix += _hex[length >> 12 & 0xf]
          prefix += _hex[length >> 8  & 0xf]
          prefix += _hex[length >> 4 & 0xf]
          prefix += _hex[length & 0xf]
          data = prefix + packet + '0000'
          data = data.encode() + p.stdout.read()
          res = make_response(data)
          res.headers['Expires'] = 'Fri, 01 Jan 1980 00:00:00 GMT'
          res.headers['Pragma'] = 'no-cache'
          res.headers['Cache-Control'] = 'no-cache, max-age=0, must-revalidate'
          res.headers['Content-Type'] = 'application/x-%s-advertisement' % service
          p.wait()
          return res
      else:
          abort(501)

  @app.route('/example/<string:project_name>/git-receive-pack', methods=('POST',))
  @app.route('/<string:project_name>/git-receive-pack', methods=('POST',))
  @auth.login_required
  def git_receive_pack(project_name):
      if project_name in available_repos:
          project_path = [v for k,v in git_repo_map.items() if k == project_name][0]
          p = subprocess.Popen(['git-receive-pack', '--stateless-rpc', project_path], stdin=subprocess.PIPE, stdout=subprocess.PIPE)
          data_in = request.data
          pack_file = data_in[data_in.index('PACK'):]
          objects = PackStreamReader(StringIO(pack_file).read)
          for obj in objects.read_objects():
              if obj.obj_type_num == 1: # Commit
                  print(obj)
          p.stdin.write(data_in)
          data_out = p.stdout.read()
          res = make_response(data_out)
          res.headers['Expires'] = 'Fri, 01 Jan 1980 00:00:00 GMT'
          res.headers['Pragma'] = 'no-cache'
          res.headers['Cache-Control'] = 'no-cache, max-age=0, must-revalidate'
          res.headers['Content-Type'] = 'application/x-git-receive-pack-result'
          p.wait()
          return res
      else:
          abort(501)

  @app.route('/example/<string:project_name>/git-upload-pack', methods=('POST',))
  @app.route('/<string:project_name>/git-upload-pack', methods=('POST',))
  @auth.login_required
  def git_upload_pack(project_name):
      if project_name in available_repos:
          project_path = [v for k,v in git_repo_map.items() if k == project_name][0]
          p = subprocess.Popen(['git-upload-pack', '--stateless-rpc', project_path],
                               stdin=subprocess.PIPE, stdout=subprocess.PIPE)
          p.stdin.write(request.data)
          p.stdin.close()
          data = p.stdout.read()
          res = make_response(data)
          res.headers['Expires'] = 'Fri, 01 Jan 1980 00:00:00 GMT'
          res.headers['Pragma'] = 'no-cache'
          res.headers['Cache-Control'] = 'no-cache, max-age=0, must-revalidate'
          res.headers['Content-Type'] = 'application/x-git-upload-pack-result'
          p.wait()
          return res
      else:
          abort(501)

  logger.info("Start API")

  app_port = args.port or app_config.get('app.port') or default_app_port
  app_host_address = args.host_address or app_config.get('app.address') or default_app_host_address
  app.run(host=app_host_address, port=app_port)

  logger.info("Stop API")

def main():
  """The main entrypoint
  """

  start_api()

if __name__ == '__main__':
  main()





import sys
if sys.version_info > (3, 8):
  from importlib import metadata
  __version__ = metadata.version("btgitserver")
else:
  import importlib_metadata;
  __version__ = importlib_metadata.version('btgitserver')


import argparse
from argparse import RawTextHelpFormatter
from btgitserver.defaults import app_name

def parse_args(**kwargs):
    parser = argparse.ArgumentParser(
        description="Standalone git server written in python",
        formatter_class=RawTextHelpFormatter,
        epilog="""
        This program is a git server implementation written in python.
        Refer to the README.md for setup and dependency information.
        The program defaults can be overridden via a yaml-formatted config file.
        or specified via commandline
        
        The config file must conform to the following data structure:
        
        auth:
          users:
            username1:
              password: password1
            username2:
              password: password2
        
        app:
          debug: false
          listen: "0.0.0.0"
          port: 5000
          search_paths: 
            - ~/repos
        """,
        fromfile_prefix_chars='@')
    parser.add_argument(
        "-host",
        "--host-address",
        help="Override listening address",
        metavar="ARG", required=False)
    parser.add_argument(
        "-p",
        "--port",
        help="Override listening port",
        metavar="ARG", required=False)
    parser.add_argument(
        "-w",
        "--workers",
        help="Override number of gunicorn workers",
        metavar="ARG", required=False)    
    parser.add_argument('--no-verify-tls', '-notls',action='store_true', help='Verify SSL cert when downloading web content')
    parser.add_argument('--config-file', '-f', help="Path to config file override")
    parser.add_argument('--repo-search-paths', '-r', nargs='+', help="List of directories containing git repositories")
    parser.add_argument('--logfile-path', '-L', help="Path to logfile")
    parser.add_argument('--logfile-write-mode', '-Lw', default='w', choices=['a', 'w'], help="File mode when writing to log file, 'a' to append, 'w' to overwrite")    
    parser.add_argument('--debug', action='store_true')
    parser.add_argument(
        "-v",
        "--verbose",
        help="increase output verbosity",
        action="store_true")
    
    return parser.parse_args()


import os

# SSL Requests
default_verify_tls = True

# Flask app settings
debug = False
app_name = "Python Git Server"
default_app_port = 5000
default_config_file_name = 'config.yaml'
default_app_host_address = '0.0.0.0'
default_repo_search_paths = ['~/repos']
default_num_workers = 4
default_settings = {
  "auth": {
    "users": {
      "git-user": {
        "password": "git-password"
      }
    }
  },
  "app": {
    "debug": debug,
    "listen": default_app_host_address,
    "port": default_app_port,
    "search_paths": default_repo_search_paths,
    "workers": default_num_workers
  }
}


import configparser as ConfigParser
import os
from setuptools import setup
import sysconfig
import sys

etc_dirname = 'etc'

cfg = ConfigParser.ConfigParser()
cfg.read('setup.cfg')
my_package_name = cfg.get('metadata', 'name')

if sys.platform == 'win32':
  site_packages_path = 'scripts'
else:
  site_packages_path = 'bin'

data_files_path = site_packages_path

def tree(src):
  result = []
  data_file_path = os.path.join(data_files_path, src)
  for root, dirs, files in os.walk(src):
    for file in files:
      if os.path.sep in root:
          sub_root = root.split(os.path.sep, 1)[-1]
          file = os.path.join(sub_root, file)
      result.append(os.path.join(src, file))
  return [(data_file_path, result)]

DATA_FILES = tree(etc_dirname)

setup(data_files=DATA_FILES)


import logging
import logging.handlers
import coloredlogs
import os
import sys

class Logger:

  def __init__(self, **kwargs):
    
    env_debug_is_on = os.environ.get('btgitserver_DEBUG', '').lower() in [
    't', 'true', '1', 'on', 'y', 'yes']
    self.debug = kwargs.get('debug', False) or env_debug_is_on
    self.FORMAT_STR = "%(asctime)s %(name)s [%(levelname)s]: %(message)s"
    self.formatter = logging.Formatter(
      self.FORMAT_STR,
      datefmt='%Y-%m-%d %H:%M:%S'
    )
    self.logfile_path = kwargs.get('logfile_path')
    self.logfile_write_mode = kwargs.get('logfile_write_mode', 'a')

  def init_logger(self, name=None):
    # Setup Logging
    logger = logging.getLogger(name)
    # TODO Find a better approach to this hacky method
    if self.debug:
        logging_level = logging.DEBUG
    else:
        logging_level = logging.INFO
    logger.setLevel(logging_level)
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setFormatter(self.formatter)
    logger.addHandler(stdout_handler)
    if self.logfile_path:
        # create one handler for print and one for export
        file_handler = logging.FileHandler(self.logfile_path, self.logfile_write_mode)
        file_handler.setFormatter(self.formatter)
        logger.addHandler(file_handler)
    coloredlogs.install(logger=logger,
                        fmt=self.FORMAT_STR,
                        level=logging_level)
    return logger  